#
# Raw-App to implement back-end logic for:
# Connect 1xFC612
# Configure TJA1101 Internal Loopback
# Open Raw-Rx/Tx
# Send Receive raw frames
# Enable/Disable IDLE
# Trigger In-Out
#

from queue import Queue
import binascii

from broadway2 import Broadway, FCAdapter, RawRxFrame, RawRxHeaderTypes, HWTimestamp, SmiAccess
from broadway2 import broadway_api, PhyAccess, get_phy_access


class TJA110x:
    def __init__(self, phyAccess):
        self.phyAccess = phyAccess

    def enable_loopback(self, enable):
        if enable:
        # linkControl = Off
            reg17 = self.phyAccess.read_reg(17)
            reg17 &= ~(1<<15)
            self.phyAccess.write_reg(17, reg17)

            # loopback ON
            reg0 = self.phyAccess.read_reg(0)
            reg0 |= (1<<14)
            self.phyAccess.write_reg(0, reg0)

            # linkControl = On
            reg17 = self.phyAccess.read_reg(17)
            reg17 |= (1<<15)
            self.phyAccess.write_reg(17, reg17)
        else:
            # linkControl = Off
            reg17 = self.phyAccess.read_reg(17)
            reg17 &= ~(1<<15)
            self.phyAccess.write_reg(17, reg17)

            # loopback OFF
            reg0 = self.phyAccess.read_reg(0)
            reg0 &= ~(1<<14)
            self.phyAccess.write_reg(0, reg0)

            # linkControl = On
            reg17 = self.phyAccess.read_reg(17)
            reg17 |= (1<<15)
            self.phyAccess.write_reg(17, reg17)




class ADIN1100:
    def __init__(self, phyAccess):
        self.phyAccess = phyAccess

    def enable_loopback(self, enable):
        if enable:
            regVal = self.phyAccess.read_dev_reg(0x1E, 0x803D)
            regVal = 1
            self.phyAccess.write_dev_reg(0x1E, 0x803D, regVal)
        else:
            regVal = self.phyAccess.read_dev_reg(0x1E, 0x803D)
            regVal &= ~1
            self.phyAccess.write_dev_reg(0x1E, 0x803D, regVal)




class AppRawLoop:
    """ requires one USB Raw-Adapter"""
    def __init__(self):
        self.base = Broadway()
        self.adapterDic = {}
        for a in self.base.enumerate():
            try:
                tst = self.adapterDic[a.DevName]
            except:
                self.adapterDic[a.DevName] = []

            phyDic = {}
            
            if a.DevName == 'FC611' or a.DevName == 'FC612':
                dev = self.base.open_control(a.Inst)
                smiAccess = SmiAccess(dev)
                res = smiAccess.detect2()
                try:
                    phyDic[0] = res[0]
                except:
                    pass
                self.base.close(dev)
                self.adapterDic[a.DevName].append({'Inst': a.Inst, 'SerialNumber': a.SerialNumber, 'Name': a.FriendlyName, 'DeviceName': a.DevName, 'Phy': phyDic})

            elif a.DevName == 'FC631':
                dev = self.base.open_control(a.Inst)
                smiAccess = SmiAccess(dev)
                res = smiAccess.detect2()
                try:
                    phyDic[0] = res[0]
                except:
                    pass
                self.base.close(dev)
                self.adapterDic[a.DevName].append({'Inst': a.Inst, 'SerialNumber': a.SerialNumber, 'Name': a.FriendlyName, 'DeviceName': a.DevName, 'Phy': phyDic})
            

        self.adapter = None
        self.phyAccess = None
        self.dev = None
        self.devRx = None
        self.devTx = None

        self.hwTimestampTx = None
        self.hwTimestampRx = None
        self.hwTimestampTrig = None

        self.qTxFrame = Queue()

    def get_adapter_dic(self):
        return self.adapterDic

    def open(self, inst):
        self.adapter = self.base.get_adapter(inst)

        # verify to have right Adapter connected
        if self.adapter == None:
            return False


        self.dev = self.base.open_control(self.adapter.Inst)
        self.adapter.connect(self.dev)

        # get direct SMI (MDIO) Register access
        self.phyAccess = get_phy_access(self.adapter)

        if self.phyAccess.supportsClause45:
            try:
                phyid = (self.phyAccess.read_dev_reg(0x1E, 2) << 16) | self.phyAccess.read_dev_reg(0x1E, 3)
                self.phy = ADIN1100(self.phyAccess)
            except:
                self.phy = None
        else:
            try:
                phyid = (self.phyAccess.read_reg(2) << 16) | self.phyAccess.read_reg(3)
                self.phy = TJA110x(self.phyAccess)
            except:
                self.phy = None

        self.devRx = self.base.open_raw_rx(self.adapter.Inst, RawRxHeaderTypes.HeaderType2)

        self.devTx = self.base.open_raw_tx(self.adapter.Inst)

        broadway_api.enable_raw_rx(self.devRx)

        return True
    

    def close(self):
        res = broadway_api.enable_idle_frames(self.devRx, 0)
        self.config_phy_loopback(False)

        self.base.close(self.devTx)
        self.base.close(self.devRx)
        self.base.close(self.dev)
        self.phyAccess = None
        self.adapter = None

    def get_link_state(self):
        try:
            return self.adapter.LinkState
        except:
            return False

    def reset(self):
        self.hwTimestampTx = None
        self.hwTimestampRx = None
        self.hwTimestampTrig = None


    def control_idle(self, timeMs):

        res = broadway_api.enable_idle_frames(self.devRx, timeMs)

        return res == 0

    def control_trigger_in(self, enable):
        if enable:
            res = broadway_api.config_ext_trigger(self.devRx, 0, 1)
        else:
            res = broadway_api.config_ext_trigger(self.devRx, 0, 0)

        return res == 0

    def config_trigger_out(self, etherTypeList, pulseWidth):
        # enable Trig-Out Pin if EthType == 0x88F7 is received
        self.adapter.config_ext_trigger_out1(1, etherTypeList, pulseWidth)

    def get_rx_frames(self):
        
        resList = []
        if self.devRx == None:
            return resList

        res = broadway_api.wait_raw_rx_frame(self.devRx, 0)

        while True:
            res = broadway_api.get_raw_rx_frame_ext(self.devRx)

            if res != None:
                rawRx = RawRxFrame(res[0], res[1])

                resList.append(rawRx)

                # Rx-Frame
                if rawRx.FrameType == 1:
                    self.hwTimestampRx = HWTimestamp(rawRx.TimestampSec, rawRx.TimestampNano)

                # Trigger-In
                if rawRx.FrameType == 3:
                    self.hwTimestampTrig = HWTimestamp(rawRx.TimestampSec, rawRx.TimestampNano)

            else:
                break

        return resList

    def get_tx_frames(self):
        resList = []

        while True:
            try:
                item = self.qTxFrame.get(False, 0)
            except:
                break

            resList.append(item)

        return resList

    def send_tx_frame(self, frameData):
        """send one frame and return (frameId, timestamp)"""
        if self.devTx == None:
            return

        res = broadway_api.wait_for_tx(self.devTx, 1000)
        if not res:
            return

        binData = binascii.a2b_hex(frameData)

        frameId = broadway_api.submit_raw_frame(self.devTx, binData)

        res = broadway_api.send_raw_frames(self.devTx)

        res = broadway_api.wait_for_raw_tx_completion(self.devTx, frameId, 1000)

        timeTx = broadway_api.get_raw_tx_timestamp(self.devTx, frameId)
        self.hwTimestampTx = HWTimestamp(timeTx[0], timeTx[1])
        self.TxFrameId = frameId

        self.qTxFrame.put((frameId, self.hwTimestampTx, frameData))

        return True

    def get_last_tx_time(self):
        return self.hwTimestampTx

    def get_last_rx_diff_time(self):
        try:
            return self.hwTimestampRx - self.hwTimestampTx
        except:
            return HWTimestamp(0,0)

    def get_last_trig_diff_time(self):
        try:
            return self.hwTimestampTrig - self.hwTimestampRx
        except:
            return HWTimestamp(0,0)


    def send_tx_frames(self, lstFrames):
        if self.devTx == None:
            return

        res = broadway_api.wait_for_tx(self.devTx, 1000)
        if not res:
            return

        cnt = 0
        for frameData in lstFrames:
            binData = binascii.a2b_hex(frameData)

            frameId = broadway_api.submit_raw_frame(self.devTx, binData)
            cnt += 1

        res = broadway_api.send_raw_frames(self.devTx)

        return cnt


    def  get_stick_time(self):
        if self.adapter == None:
            return '0000.0000'
        return self.adapter.get_timestamp()

    def get_mac_address(self):
        return self.adapter.MACAddress


    def config_phy_loopback(self, enableLoopback):
        """uses phy internal loopback Mode"""

        if self.phyAccess == None:
            return False
        if self.phy == None:
            return False

        self.phy.enable_loopback(enableLoopback)

       


    def force_trigger_out(self):
        if self.phyAccess == None:
            return False

        self.adapter.force_trigger_out1()
        return True






